import os
from os import path
from collections import defaultdict
try:
    from ConfigParser import ConfigParser  # pragma: no cover
except ImportError:  # pragma: no cover
    from configparser import ConfigParser  # pragma: no cover

from .const import API_PATH, CONFIG_FILE_NAME, DEFAULT_CACHE_TIMEOUT, ENVs, GET, POST
from .exceptions import TRAWLoginError
from . import models
from .sessions import Session
from .utils import cacheable, cacheable_generator, clear_cache, paginate

_USER_KEY = 'username'
_PASS_KEY = 'password'
_URL_KEY = 'url'


class API(object):
    """ TRAW's interface class to TestRail's REST API

    The API class is not meant to be accessed directly, rather, use the traw.Client
    """
    cache_timeouts = defaultdict(lambda: defaultdict(lambda: DEFAULT_CACHE_TIMEOUT))

    def __init__(self, username=None, user_api_key=None, password=None, url=None):
        """
        """
        config = _load_config()
        _username = username or _env_var(_USER_KEY) or config[_USER_KEY]
        _password = user_api_key or password or _env_var(_PASS_KEY) or config[_PASS_KEY]
        _url = url or _env_var(_URL_KEY) or config[_URL_KEY]

        if _username is None or _password is None or _url is None:
            msg = ('You must set a username, password/api_key, and url to '
                   'use TRAW')
            raise TRAWLoginError(msg)

        self._session = Session(auth=(_username, _password), url=_url)

    @cacheable(models.Case)
    def case_by_id(self, case_id):
        """ Calls `get_case` API endpoint with the given case_id

        :param case_id: int id of case

        :returns: case dict
        """
        path = API_PATH['get_case'].format(case_id=case_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Case)
    def cases_by_project_id(self, project_id, **params):
        """ Calls `get_cases` API endpoint

        :yields: case dictionaries from api
        """
        path = API_PATH['get_cases'].format(project_id=project_id)
        for case in self._session.request(method=GET, path=path, params=params):
            yield case

    @cacheable_generator(models.CaseType)
    def case_types(self):
        """ Calls `get_case_types` API endpoint

        :yields: case_type dictionaries from api
        """
        path = API_PATH['get_case_types']
        for case_type in self._session.request(method=GET, path=path):
            yield case_type

    @cacheable_generator(models.ConfigGroup)
    def config_groups(self, project_id):
        """ Calls `get_configs` API endpoint

        :yields: config_group dictionaries from api
        """
        path = API_PATH['get_configs'].format(project_id=project_id)

        for config_group in self._session.request(method=GET, path=path):
            yield config_group

    @clear_cache(config_groups)
    def config_add(self, config_group_id, params):
        path = API_PATH['add_config'].format(config_group_id=config_group_id)
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(config_groups)
    def config_delete(self, config_id):
        path = API_PATH['delete_config'].format(config_id=config_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(config_groups)
    def config_update(self, config_id, params):
        path = API_PATH['update_config'].format(config_id=config_id)
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(config_groups)
    def config_group_add(self, project_id, params):
        path = API_PATH['add_config_group'].format(project_id=project_id)
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(config_groups)
    def config_group_delete(self, config_group_id):
        path = API_PATH['delete_config_group'].format(config_group_id=config_group_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(config_groups)
    def config_group_update(self, config_group_id, params):
        path = API_PATH['update_config_group'].format(config_group_id=config_group_id)
        return self._session.request(method=POST, path=path, json=params)

    @cacheable(models.Milestone)
    def milestone_by_id(self, milestone_id):
        """ Calls `get_milestone` API endpoint with the given milestone_id

        :param milestone_id: int id of milestone

        :returns: milestone dict
        """
        path = API_PATH['get_milestone'].format(milestone_id=milestone_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Milestone)
    def milestones(self, project_id, is_completed=None, is_started=None):
        """ Calls `get_milestones` API endpoint

        :yields: milestone dictionaries from api
        """
        is_comp = int(is_completed) if is_completed is not None else None
        is_star = int(is_started) if is_started is not None else None

        path = API_PATH['get_milestones'].format(project_id=project_id)

        params = dict()
        if is_completed is not None:
            params['is_completed'] = is_comp

        if is_started is not None:
            params['is_started'] = is_star

        params = params or None

        for milestone in self._session.request(method=GET, path=path, params=params):
            yield milestone

    @clear_cache(milestones)
    @clear_cache(milestone_by_id)
    def milestone_add(self, project_id, params):
        path = API_PATH['add_milestone'].format(project_id=project_id)
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(milestones)
    @clear_cache(milestone_by_id)
    def milestone_delete(self, milestone_id):
        path = API_PATH['delete_milestone'].format(milestone_id=milestone_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(milestones)
    @clear_cache(milestone_by_id)
    def milestone_update(self, milestone_id, params):
        path = API_PATH['update_milestone'].format(milestone_id=milestone_id)
        return self._session.request(method=POST, path=path, json=params)

    @cacheable(models.Plan)
    def plan_by_id(self, plan_id):
        """ Calls `get_plan` API endpoint with the given plan_id

        :param plan_id: int id of plan

        :returns: plan dict
        """
        path = API_PATH['get_plan'].format(plan_id=plan_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Priority)
    def priorities(self):
        """ Calls `get_priorities` API endpoint

        :yields: priority dictionaries from api
        """
        path = API_PATH['get_priorities']
        for priority in self._session.request(method=GET, path=path):
            yield priority

    @cacheable(models.Project)
    def project_by_id(self, project_id):
        """ Calls `get_project` API endpoint with the given project_id

        :param project_id: int id of project

        :returns: project dict
        """
        path = API_PATH['get_project'].format(project_id=project_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Project)
    def projects(self, is_completed=None):
        """ Calls `projects` API endpoint with given filter

        :param project_filter: Filter results by completion status: 0, 1, or None

        :yields: project dictionaries from api
        """
        path = API_PATH['get_projects']
        if is_completed is not None:
            params = dict(is_completed=is_completed)
        else:
            params = None

        for project in self._session.request(method=GET, path=path, params=params):
            yield project

    @clear_cache(projects)
    @clear_cache(project_by_id)
    def project_add(self, params):
        path = API_PATH['add_project']
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(projects)
    @clear_cache(project_by_id)
    def project_delete(self, project_id):
        path = API_PATH['delete_project'].format(project_id=project_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(projects)
    @clear_cache(project_by_id)
    def project_update(self, project_id, params):
        path = API_PATH['update_project'].format(project_id=project_id)
        return self._session.request(method=POST, path=path, json=params)

    @cacheable_generator(models.Result)
    @paginate
    def results_by_test_id(self, test_id, **params):
        """ Calls `get_results` API endpoint

        :yields: result dictionaries from api
        """
        path = API_PATH['get_results'].format(test_id=test_id)
        for result in self._session.request(method=GET, path=path, params=params):
            yield result

    @clear_cache(results_by_test_id)
    def result_add(self, test_id, params):
        path = API_PATH['add_result'].format(test_id=test_id)
        return self._session.request(method=POST, path=path, json=params)

    @cacheable(models.Run)
    def run_by_id(self, run_id):
        """ Calls `get_run` API endpoint with the given run_id

        :param run_id: int id of run

        :returns: run dict
        """
        path = API_PATH['get_run'].format(run_id=run_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Run)
    @paginate
    def runs_by_project_id(self, project_id, **params):
        """ Calls `get_runs` API endpoint

        :yields: run dictionaries from api
        """
        path = API_PATH['get_runs'].format(project_id=project_id)

        for run in self._session.request(method=GET, path=path, params=params):
            yield run

    @clear_cache(run_by_id)
    def run_add(self, project_id, params):
        path = API_PATH['add_run'].format(project_id=project_id)
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(run_by_id)
    def run_close(self, run_id):
        path = API_PATH['close_run'].format(run_id=run_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(run_by_id)
    def run_delete(self, run_id):
        path = API_PATH['delete_run'].format(run_id=run_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(run_by_id)
    def run_update(self, run_id, params):
        path = API_PATH['update_run'].format(run_id=run_id)
        return self._session.request(method=POST, path=path, json=params)

    @cacheable(models.Section)
    def section_by_id(self, section_id):
        """ Calls `get_section` API endpoint with the given section_id

        :param section_id: int id of section

        :returns: section dict
        """
        path = API_PATH['get_section'].format(section_id=section_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Section)
    def sections_by_project_id(self, project_id, suite_id=None):
        """ Calls `get_sections` API endpoint

        :yields: section dictionaries from api
        """
        params = {'suite_id': suite_id} if suite_id else dict()
        path = API_PATH['get_sections'].format(project_id=project_id)
        for section in self._session.request(method=GET, path=path, params=params):
            yield section

    @clear_cache(section_by_id)
    @clear_cache(sections_by_project_id)
    def section_add(self, project_id, params):
        path = API_PATH['add_section'].format(project_id=project_id)
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(section_by_id)
    @clear_cache(sections_by_project_id)
    def section_delete(self, section_id):
        path = API_PATH['delete_section'].format(section_id=section_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(section_by_id)
    @clear_cache(sections_by_project_id)
    def section_update(self, section_id, params):
        path = API_PATH['update_section'].format(section_id=section_id)
        return self._session.request(method=POST, path=path, json=params)

    @cacheable_generator(models.Status)
    def statuses(self):
        """ Calls `get_statuses` API endpoint

        :yields: status dictionaries from api
        """
        path = API_PATH['get_statuses']
        for status in self._session.request(method=GET, path=path):
            yield status

    @cacheable(models.Suite)
    def suite_by_id(self, suite_id):
        """ Calls `get_suite` API endpoint with the given suite_id

        :param suite_id: int id of suite

        :returns: suite dict
        """
        path = API_PATH['get_suite'].format(suite_id=suite_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Suite)
    def suites_by_project_id(self, project_id):
        """ Calls `get_suites` API endpoint

        :yields: suite dictionaries from api
        """
        path = API_PATH['get_suites'].format(project_id=project_id)
        for suite in self._session.request(method=GET, path=path):
            yield suite

    @clear_cache(suite_by_id)
    @clear_cache(suites_by_project_id)
    def suite_add(self, project_id, params):
        path = API_PATH['add_suite'].format(project_id=project_id)
        return self._session.request(method=POST, path=path, json=params)

    @clear_cache(suite_by_id)
    @clear_cache(suites_by_project_id)
    def suite_delete(self, suite_id):
        path = API_PATH['delete_suite'].format(suite_id=suite_id)
        return self._session.request(method=POST, path=path)

    @clear_cache(suite_by_id)
    @clear_cache(suites_by_project_id)
    def suite_update(self, suite_id, params):
        path = API_PATH['update_suite'].format(suite_id=suite_id)
        return self._session.request(method=POST, path=path, json=params)

    @cacheable_generator(models.Template)
    def templates(self, project_id):
        """ Calls `get_templates` API endpoint

        :yields: template dictionaries from api
        """
        path = API_PATH['get_templates'].format(project_id=project_id)
        for template in self._session.request(method=GET, path=path):
            yield template

    @cacheable(models.Test)
    def test_by_id(self, test_id):
        """ Calls `get_test` API endpoint with the given test_id

        :param test_id: int id of test

        :returns: test dict
        """
        path = API_PATH['get_test'].format(test_id=test_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.Test)
    def tests_by_run_id(self, run_id, status_id=None):
        """ Calls `get_tests` API endpoint

        :yields: test dictionaries from api
        """
        path = API_PATH['get_tests'].format(run_id=run_id)
        params = {'status_id': status_id} if status_id else None
        for test in self._session.request(method=GET, path=path, params=params):
            yield test

    @cacheable(models.User)
    def user_by_email(self, email):
        """ Calls `get_user` API endpoint with the given user email

        :param email: str email of user

        :returns: user dict
        """
        path = API_PATH['get_user_by_email']
        params = {'email': email}
        return self._session.request(method=GET, path=path, params=params)

    @cacheable(models.User)
    def user_by_id(self, user_id):
        """ Calls `get_user` API endpoint with the given user_id

        :param user_id: int id of user

        :returns: user dict
        """
        path = API_PATH['get_user'].format(user_id=user_id)
        return self._session.request(method=GET, path=path)

    @cacheable_generator(models.User)
    def users(self):
        """ Calls `users` API endpoint

        :yields: user dictionaries from api
        """
        path = API_PATH['get_users']
        for user in self._session.request(method=GET, path=path):
            yield user


def _env_var(var_name):
    """ Load the target environment variable, if it exists

    :param var_name: Name of the environment variable to load

    :returns: string if the env var exists, None if it does not
    """
    var = None

    if var_name == _USER_KEY:
        var = os.environ.get(ENVs.USER_KEY, None)
    elif var_name == _PASS_KEY:
        var = os.environ.get(ENVs.API_KEY, None) or os.environ.get(ENVs.PASS_KEY, None)
    elif var_name == _URL_KEY:
        var = os.environ.get(ENVs.URL_KEY, None)
    else:
        msg = "'var_name' must be {0}, {1}, or {2}"
        raise ValueError(msg.format(_USER_KEY, _PASS_KEY, _URL_KEY))

    return var


def _load_config():
    """ Load the configuration file, if it exists

    :returns: dict
    """
    config_dict = {_USER_KEY: None, _PASS_KEY: None, _URL_KEY: None}

    conf_file_path = path.expanduser(path.join('~', CONFIG_FILE_NAME))
    if path.exists(conf_file_path):
        config_parser = ConfigParser()
        config_parser.read(conf_file_path)
        t_conf = config_parser['TRAW']

        config_dict[_USER_KEY] = t_conf.get('username', None)
        config_dict[_PASS_KEY] = t_conf.get('user_api_key', None) or t_conf.get('password', None)
        config_dict[_URL_KEY] = t_conf.get('url', None)

    return config_dict



from .model_base import ModelBase


class Template(ModelBase):
    """ Object model for TestRail Templates

    To get all templates for a project:

    .. code-block:: python

        target_project = traw_client.project(12)  # Get project with id 12
        templates_for_project = list(traw_client.template(target_project))

    """
    @property
    def is_default(self):
        """ True if this is a default template """
        return self._content.get('is_default')

    @property
    def name(self):
        """ The name of the template """
        return self._content.get('name')



from datetime import datetime as dt

from .model_base import ModelBase
from ..const import PROJECT_ADD_FIELDS, PROJECT_UPDATE_FIELDS, SETTER_ERR
from .posters import Addable, Deleteable, Updatable


class Project(Addable, Deleteable, Updatable, ModelBase):
    """ Object model for TestRail Projects

    To create new project

    .. code-block:: python

        new_project = traw_client.project()
        new_project.name = "My new TestRail project"
        new_project.announcement = "My new project announcment"
        new_project.show_announcement = True
        new_project.suite_mode = 1

    """
    _ADDABLE_FIELDS = PROJECT_ADD_FIELDS
    _UPDATABLE_FIELDS = PROJECT_UPDATE_FIELDS

    @property
    def announcement(self):
        """ The description/announcement of the project
            Can be set to any string
        """
        return self._content.get('announcement', None)

    @announcement.setter
    def announcement(self, val):
        if not isinstance(val, str):
            raise TypeError(SETTER_ERR.format(str, type(val)))
        self._content['announcement'] = val

    @property
    def completed_on(self):
        """ The date/time when the project was marked as completed
            (as datetime.datetime object)
        """
        co = self._content.get('completed_on', None)
        if co:
            co = dt.fromtimestamp(co)
        return co

    @property
    def is_completed(self):
        """ True if the project is marked as completed and false otherwise """
        return self._content.get('is_completed', False)

    @is_completed.setter
    def is_completed(self, val):
        if not isinstance(val, bool):
            raise TypeError(SETTER_ERR.format(bool, type(val)))
        self._content['is_completed'] = val

    @property
    def name(self):
        """ The name of the project. Can be set to any string """
        return self._content.get('name', None)

    @name.setter
    def name(self, val):
        if not isinstance(val, str):
            raise TypeError(SETTER_ERR.format(str, type(val)))
        self._content['name'] = val

    @property
    def show_announcement(self):
        """ To show or not show an announement. Can be set to True or False """
        return self._content.get('show_announcement', False)

    @show_announcement.setter
    def show_announcement(self, val):
        if not isinstance(val, bool):
            raise TypeError(SETTER_ERR.format(bool, type(val)))
        self._content['show_announcement'] = val

    @property
    def suite_mode(self):
        """ The suite mode of the project (1 for single suite mode,
            2 for single suite + baselines, 3 for multiple suites)
            (added with TestRail 4.0)
        """
        return self._content.get('suite_mode', None)

    @suite_mode.setter
    def suite_mode(self, val):
        if not isinstance(val, int):
            raise TypeError(SETTER_ERR.format(int, type(val)))
        elif val not in [1, 2, 3]:
            raise ValueError('suite_mode can only be set to 1, 2, or 3')
        self._content['suite_mode'] = val

    @property
    def url(self):
        """ The address/URL of the project in the user interface """
        return self._content.get('url', None)



from __future__ import print_function

import traw


def main():
    client = traw.Client()  # Credentials loaded through ENV vars

    # Tests are retrieved by the run they are associated with
    # Get an example run using the Run ID
    auto_run = client.run(72215)

    # Get all tests for this run:
    tests = list(client.tests(auto_run))  # Expand the generator to a list
    print(len(tests))  # 40 tests

    test_1 = tests[0]
    print(test_1.title)  # Example Test Case Title
    print(test_1.status.label)  # Passed

    test_2 = tests[1]
    print(test_2.title)  # Example Test Case Title #2
    print(test_2.status.label)  # Failed

    # Get all tests with a status of `Failed` or 'Retest'
    failed_status = client.status('failed')
    retest_status = client.status('retest')
    failed_and_retest = list(client.tests(auto_run, with_status=[failed_status, retest_status]))

    print(len(failed_and_retest))  # 10

    # Get a test from its test id
    example_test = client.test(5792090)
    print(example_test.id)  # 5792090
    print(example_test.status.label)  # Failed
    print(example_test.type.name)  # Regression
    print(example_test.priority.name)  # Medium
    print(example_test.estimate)  # datetime.timedelta object -> datetime.timedelta(0, 4200)
    print(list(example_test.refs))  # ['REF001', 'REF002']

    # Since you cannot programatically add Tests through the API,
    # attempting to call `client.test()` will raise an exception rather
    # than returning an empty/new Test object
    try:
        client.test()
    except NotImplementedError as exc:
        # Not implemented directly. You must pass in int object
        print(str(exc))


if __name__ == "__main__":
    main()



from copy import deepcopy

import pytest

from traw.models import Config, ConfigGroup, Project

CONFIG_GROUP_ID = 555
CONFIG_GROUP_NAME = 'mock config group name'
CONFIG_GROUP_NAME_NEW = 'new mock config group name'
CONFIG_ID = 666
CONFIG_NAME = 'mock config name'
CONFIG_NAME_NEW = "new config name"
PROJECT_ID = 15
CONFIG_CONTENT = {"group_id": CONFIG_GROUP_ID,
                  "id": CONFIG_ID,
                  "name": CONFIG_NAME}
CONFIG_GROUP_CONTENT = {"configs": [CONFIG_CONTENT],
                        "id": CONFIG_GROUP_ID,
                        "name": CONFIG_GROUP_NAME,
                        "project_id": PROJECT_ID}


@pytest.fixture()
def empty_config(client):
    return Config(client)


@pytest.fixture()
def empty_config_group(client):
    return ConfigGroup(client)


@pytest.fixture()
def config(client):
    return Config(client, deepcopy(CONFIG_CONTENT))


@pytest.fixture()
def config_w_project(client):
    project = Project(client, {'id': PROJECT_ID})
    return Config(client, deepcopy(CONFIG_CONTENT), project)


@pytest.fixture()
def config_group(client):
    return ConfigGroup(client, deepcopy(CONFIG_GROUP_CONTENT))


def test_config_project(empty_config, config, config_w_project):
    config.client.api.project_by_id.return_value = {'id': PROJECT_ID}

    assert empty_config.project is None

    assert config.project is None

    assert isinstance(config_w_project.project, Project)
    assert config_w_project.project.id == PROJECT_ID
    config.client.api.project_by_id.assert_called_with(PROJECT_ID)


def test_config_project_set(empty_config):
    project = Project(None, {'id': PROJECT_ID})

    assert empty_config.project is None

    empty_config.project = project

    assert empty_config._content['project_id'] == PROJECT_ID


def test_config_project_set_exc(empty_config):
    with pytest.raises(TypeError) as exc:
        empty_config.project = 1234

    assert str(Project) in str(exc)


def test_config_name_get(empty_config, config, config_w_project):
    assert empty_config.name is None
    assert config.name == CONFIG_NAME
    assert config_w_project.name == CONFIG_NAME


def test_config_name_set(empty_config, config, config_w_project):
    assert empty_config.name is None
    empty_config.name = CONFIG_NAME_NEW
    assert empty_config.name == CONFIG_NAME_NEW

    assert config.name == CONFIG_NAME
    config.name = CONFIG_NAME_NEW
    assert config.name == CONFIG_NAME_NEW

    assert config_w_project.name == CONFIG_NAME
    config_w_project.name = CONFIG_NAME_NEW
    assert config_w_project.name == CONFIG_NAME_NEW


def test_config_name_set_exc(empty_config):
    with pytest.raises(TypeError) as exc:
        empty_config.name = 1234

    assert str(str) in str(exc)
    assert str(int) in str(exc)


def test_config_config_group_get(empty_config, config, config_w_project):
    config.client.api.project_by_id.return_value = {'id': PROJECT_ID}
    config.client.api.config_groups.return_value = [CONFIG_GROUP_CONTENT]

    assert empty_config.config_group is None
    assert empty_config.client.api.method_calls == list()
    assert not empty_config.client.api.project_by_id.called
    assert not empty_config.client.api.config_groups.called

    assert isinstance(config.config_group, ConfigGroup)
    assert config.config_group.id == CONFIG_GROUP_ID
    assert config.config_group.name is None
    assert not config.client.api.project_by_id.called
    assert not config.client.api.config_groups.called

    assert isinstance(config_w_project.config_group, ConfigGroup)
    assert config_w_project.config_group.id == CONFIG_GROUP_ID
    assert config_w_project.config_group.name == CONFIG_GROUP_NAME
    config_w_project.client.api.project_by_id.assert_called_with(PROJECT_ID)
    config_w_project.client.api.config_groups.assert_called_with(PROJECT_ID)


def test_config_config_group_set(empty_config):
    assert empty_config.config_group is None
    empty_config.config_group = ConfigGroup(None, {'id': CONFIG_GROUP_ID})
    assert empty_config.config_group.id == CONFIG_GROUP_ID


def test_config_config_group(empty_config):
    with pytest.raises(TypeError) as exc:
        empty_config.config_group = 1234

    assert str(ConfigGroup) in str(exc)
    assert str(int) in str(exc)


def test_config_group_configs(empty_config_group, config_group):
    empty_config_group.client.api.project_by_id.return_value = {'id': PROJECT_ID}
    config_group.client.api.project_by_id.return_value = {'id': PROJECT_ID}

    assert list(empty_config_group.configs) == list()
    assert not empty_config_group.client.api.project_by_id.called

    assert len(list(config_group.configs)) == 1
    assert isinstance(next(config_group.configs), Config)
    assert next(config_group.configs).id == CONFIG_ID
    assert next(config_group.configs).name == CONFIG_NAME
    assert config_group.client.api.project_by_id.called



from copy import deepcopy
from datetime import timedelta

import pytest

from traw.models import Test as _Test  # To avoid tox test collection warning
from traw.models import Case, CaseType, Milestone, Priority, Run, Status, User

CASE_ID = 11
CUSTOM_EXP_COND = "mock custom expected condition"
CUSTOM_PRECOND = "mock custom precondition"
ID = 123
MILESTONE_ID = 66
PRIORITY_ID = 22
REF1 = 'REF1'
REF2 = 'REF2'
REF3 = 'REF3'
REF4 = 'REF4'
RUN_ID = 33
STATUS_ID = 44
TITLE = "mock test title"
TYPE_ID = 55
USER_ID = 99


@pytest.fixture()
def empty_test(client):
    return _Test(client)


@pytest.fixture()
def test(client):
    content = {"assignedto_id": USER_ID,
               "case_id": CASE_ID,
               "custom_expected": CUSTOM_EXP_COND,
               "custom_preconds": CUSTOM_PRECOND,
               "custom_steps_separated": [
                   {
                       "content": "Step 1",
                       "expected": "Expected Result 1"
                   },
                   {
                       "content": "Step 2",
                       "expected": "Expected Result 2"
                   }
               ],
               "estimate": "1m 5s",
               "estimate_forecast": "10d 11h",
               "id": ID,
               "milestone_id": MILESTONE_ID,
               "priority_id": PRIORITY_ID,
               "refs": ",".join([REF1, REF2, REF3, REF4]),
               "run_id": RUN_ID,
               "status_id": STATUS_ID,
               "title": TITLE,
               "type_id": TYPE_ID}
    return _Test(client, deepcopy(content))


def test_assigned_to(empty_test, test):
    assert empty_test.assigned_to is None

    test.client.api.user_by_id.return_value = {'id': USER_ID}

    assert isinstance(test.assigned_to, User)
    assert test.assigned_to.id == USER_ID
    test.client.api.user_by_id.assert_called_with(USER_ID)


def test_case(empty_test, test):
    with pytest.raises(KeyError) as exc:
        empty_test.case

    assert 'case_id' in str(exc)

    test.client.api.case_by_id.return_value = {'id': CASE_ID}

    assert isinstance(test.case, Case)
    assert test.case.id == CASE_ID
    test.client.api.case_by_id.assert_called_with(CASE_ID)


def test_estimate(empty_test, test):
    assert empty_test.estimate is None
    assert isinstance(test.estimate, timedelta)
    assert test.estimate == timedelta(seconds=65)


def test_estimate_forecast(empty_test, test):
    assert empty_test.estimate_forecast is None
    assert isinstance(test.estimate_forecast, timedelta)
    assert test.estimate_forecast == timedelta(days=10, seconds=39600)


def test_milestone(empty_test, test):
    assert empty_test.milestone is None

    test.client.api.milestone_by_id.return_value = {'id': MILESTONE_ID}

    assert isinstance(test.milestone, Milestone)
    assert test.milestone.id == MILESTONE_ID
    test.client.api.milestone_by_id.assert_called_with(MILESTONE_ID)


def test_priority(empty_test, test):
    with pytest.raises(KeyError) as exc:
        empty_test.priority

    assert 'priority_id' in str(exc)

    test.client.api.priorities.return_value = [{'id': 998}, {'id': PRIORITY_ID}, {'id': 999}]

    assert isinstance(test.priority, Priority)
    assert test.priority.id == PRIORITY_ID
    assert test.client.api.priorities.call_count == 2


def test_refs(empty_test, test):
    assert list(empty_test.refs) == list()
    assert len(list(test.refs)) == 4
    assert all(map(lambda x: isinstance(x, str), test.refs))
    assert REF1 in test.refs
    assert REF2 in test.refs
    assert REF3 in test.refs
    assert REF4 in test.refs


def test_run(empty_test, test):
    with pytest.raises(KeyError) as exc:
        empty_test.run

    assert 'run_id' in str(exc)

    test.client.api.run_by_id.return_value = {'id': RUN_ID}

    assert isinstance(test.run, Run)
    assert test.run.id == RUN_ID
    test.client.api.run_by_id.assert_called_with(RUN_ID)


def test_status(empty_test, test):
    with pytest.raises(KeyError) as exc:
        empty_test.status

    assert 'status_id' in str(exc)

    test.client.api.statuses.return_value = [{'id': 998}, {'id': STATUS_ID}, {'id': 999}]

    assert isinstance(test.status, Status)
    assert test.status.id == STATUS_ID
    assert test.client.api.statuses.call_count == 2


def test_title(empty_test, test):
    with pytest.raises(KeyError) as exc:
        empty_test.title

    assert 'title' in str(exc)

    assert isinstance(test.title, str)
    assert test.title == TITLE


def test_type(empty_test, test):
    with pytest.raises(KeyError) as exc:
        empty_test.type

    assert 'type_id' in str(exc)

    test.client.api.case_types.return_value = [{'id': 998}, {'id': TYPE_ID}, {'id': 999}]

    assert isinstance(test.type, CaseType)
    assert test.type.id == TYPE_ID
    assert test.client.api.case_types.call_count == 2



from __future__ import print_function

import traw
from traw.exceptions import TRAWClientError


def main():
    client = traw.Client()  # Credentials loaded through ENV vars

    # TestRail returns config groups based on project, with each config group
    # containing specific configs. To get a config group, you must first get
    # the corresponding project by project ID:
    automation_project = client.project(15)

    # Get all config groups for our project. This method returns a generator, so
    # Use `list` to get a complete list of config groups
    config_groups = list(client.config_groups(automation_project))
    print(len(config_groups))  # 1

    config_group = config_groups[0]
    print(str(config_group))  # ConfigGroup-4
    print(config_group.name)  # 'Supported Browsers'
    print(config_group.project)  # Project-15
    print(len(list(config_group.configs)))  # 2 configs

    # Get/examine configs for each config group
    configs = config_groups[0].configs
    config_1 = next(configs)
    print(config_1.name)  # Chrome
    print(config_1.config_group)  # ConfigGroup-4, a reference back to the CG
    print(config_1.project)  # Project-15, a reference back to the project

    config_2 = next(configs)
    print(config_2.name)  # Firefox
    print(config_2.config_group)  # ConfigGroup-4, a reference back to the CG
    print(config_2.project)  # Project-15, a reference back to the project

    # You can retrieve a ConfigGroup  using project + config group ID
    print(client.config_group(automation_project, 4))  # ConfigGroup-4

    # You can create new config groups and configs and add them to TestRail
    new_config_group = client.config_group()  # Returns a new/empty config group

    # Set the name and associate it with a project
    new_config_group.name = "Supported Operating Systems"
    new_config_group.project = automation_project

    # Add the new config group to TestRail, and collect the returned object
    created_config_group = client.add(new_config_group)

    # Note that the returned config group matches the one you added and now has
    # and ID, but no associated configs
    print(created_config_group.name == new_config_group.name)  # True
    print(created_config_group.id)  # 5
    print(list(created_config_group.configs))  # []

    # Create new configs, associate them with a config group, and add them to TR
    new_config_1 = client.config()
    new_config_1.name = "Windows"
    new_config_1.config_group = created_config_group
    created_config_1 = client.add(new_config_1)
    print(created_config_1)  # Windows

    new_config_2 = client.config()
    new_config_2.name = "Linux"
    new_config_2.config_group = created_config_group
    created_config_2 = client.add(new_config_2)
    print(created_config_2)  # Linux

    new_config_3 = client.config()
    new_config_3.name = "mac OS"
    new_config_3.config_group = created_config_group
    created_config_3 = client.add(new_config_3)
    print(created_config_3)  # mac OS

    # Re-acquire the config group from the TestRail API
    updated_config_group = client.config_group(created_config_group.project,
                                               created_config_group.id)

    # Note that it now has three configs:
    cg_configs = list(updated_config_group.configs)
    print(cg_configs)  # [Config-25, Config-26, Config-24]

    print(cg_configs[0].name)  # Linux
    print(cg_configs[1].name)  # mac OS
    print(cg_configs[2].name)  # Windows

    # Should you need to update a config group or config
    to_be_updated_config = cg_configs[0]
    to_be_updated_config.name = "Linux: Ubuntu"  # From 'Linux' to 'Linux: Ubuntu'
    updated_config = client.update(to_be_updated_config)
    print(updated_config.name)  # 'Linux: Ubuntu'

    # Re-acquire the config group from the TestRail API
    updated_config_group = client.config_group(updated_config_group.project,
                                               updated_config_group.id)

    cg_configs = updated_config_group.configs
    print(next(cg_configs).name)  # Linux: Ubuntu
    print(next(cg_configs).name)  # mac OS
    print(next(cg_configs).name)  # Windows

    # Finally, should you want to delete a config/config group
    client.delete(updated_config_group)

    # And it is no longer in TestRail
    try:
        client.config_group(updated_config_group.project, updated_config_group.id)
    except TRAWClientError as exc:
        # Could not locate a models.ConfigGroup with id of 4 for project with ID 15
        print(str(exc))


if __name__ == "__main__":
    main()



class ENVs(object):
    """ OS Envionment variable names """
    USER_KEY = 'TRAW_USERNAME'
    API_KEY = 'TRAW_USER_API_KEY'
    PASS_KEY = 'TRAW_PASSWORD'
    URL_KEY = 'TRAW_URL'


API_PATH = {
    'add_case': 'add_case/{seciont_id}',
    'add_config': 'add_config/{config_group_id}',
    'add_config_group': 'add_config_group/{project_id}',
    'add_milestone': 'add_milestone/{project_id}',
    'add_plan': 'add_plan/{project_id}',
    'add_plan_entry': 'add_plan_entry/{plan_id}',
    'add_project': 'add_project',
    'add_result': 'add_result/{test_id}',
    'add_result_for_case': 'add_result_for_case/{run_id}/{case_id}',
    'add_results': 'add_results/{run_id}',
    'add_results_for_cases': 'add_results_for_cases/{run_id}',
    'add_run': 'add_run/{project_id}',
    'add_section': 'add_section/{project_id}',
    'add_suite': 'add_suite/{project_id}',
    'close_plan': 'close_plan/{plan_id}',
    'close_run': 'close_run/{run_id}',
    'delete_case': 'delete_case/{case_id}',
    'delete_config': 'delete_config/{config_id}',
    'delete_config_group': 'delete_config_group/{config_group_id}',
    'delete_milestone': 'delete_milestone/{milestone_id}',
    'delete_plan': 'delete_plan/{plan_id}',
    'delete_plan_entry': 'delete_plan_entry/{plan_id}/{entry_id}',
    'delete_project': 'delete_project/{project_id}',
    'delete_run': 'delete_run/{run_id}',
    'delete_section': 'delete_section/{section_id}',
    'delete_suite': 'delete_suite/{suite_id}',
    'get_case': 'get_case/{case_id}',
    'get_cases': 'get_cases/{project_id}',
    'get_case_fields': 'get_case_fields',
    'get_case_types': 'get_case_types',
    'get_configs': 'get_configs/{project_id}',
    'get_milestone': 'get_milestone/{milestone_id}',
    'get_milestones': 'get_milestones/{project_id}',
    'get_plan': 'get_plan/{plan_id}',
    'get_plans': 'get_plans/{project_id}',
    'get_priorities': 'get_priorities',
    'get_project': 'get_project/{project_id}',
    'get_projects': 'get_projects',
    'get_results': 'get_results/{test_id}',
    'get_results_for_case': 'get_results_for_case/{run_id}/{case_id}',
    'get_results_for_run': 'get_results_for_run/{run_id}',
    'get_result_fields': 'get_result_fields',
    'get_run': 'get_run/{run_id}',
    'get_runs': 'get_runs/{project_id}',
    'get_section': 'get_section/{section_id}',
    'get_sections': 'get_sections/{project_id}',  # suite_id may be optional
    'get_statuses': 'get_statuses',
    'get_suite': 'get_suite/{suite_id}',
    'get_suites': 'get_suites/{project_id}',
    'get_templates': 'get_templates/{project_id}',
    'get_test': 'get_test/{test_id}',
    'get_tests': 'get_tests/{run_id}',
    'get_user': 'get_user/{user_id}',
    'get_user_by_email': 'get_user_by_email',
    'get_users': 'get_users',
    'update_case': 'update_case/{case_id}',
    'update_config': 'update_config/{config_id}',
    'update_config_group': 'update_config_group/{config_group_id}',
    'update_milestone': 'update_milestone/{milestone_id}',
    'update_plan': 'update_plan/{plan_id}',
    'update_plan_entry': 'update_plan_entry/{plan_id}/{entry_id}',
    'update_project': 'update_project/{project_id}',
    'update_run': 'update_run/{run_id}',
    'update_section': 'update_section/{section_id}',
    'update_suite': 'update_suite/{suite_id}',
}

BASE_API_PATH = '/index.php?/api/v2'

CONFIG_FILE_NAME = '.traw_config'

DEFAULT_CACHE_TIMEOUT = 300  # Seconds

DEFAULT_LIMIT = 250

GET = 'get'
POST = 'post'

# Session retry parameters
DELAY = 1
RETRIES = 3

# Exception messages
NOTIMP = "Not implemented directly. You must pass in {0} object"
SETTER_ERR = 'Expected {0}, found {1}'

TIMEOUT = 16  # TODO: Make this configurable

# POST param names
ANNOUNCEMENT = 'announcement'
ASSIGNEDTO_ID = 'assignedto_id'
CASE_IDS = 'case_ids'
COMMENT = 'comment'
DEFECTS = 'defects'
DESCRIPTION = 'description'
DUE_ON = 'due_on'
ELAPSED = 'elapsed'
ESTIMATE = 'estimate'
ID = 'id'
INCLUDE_ALL = 'include_all'
IS_COMPLETED = 'is_completed'
IS_STARTED = 'is_started'
MILESTONE_ID = 'milestone_id'
NAME = 'name'
PARENT_ID = 'parent_id'
PRIORITY_ID = 'priority_id'
PROJECT_ID = 'project_id'
REFS = 'refs'
SHOW_ANNOUNCEMENT = 'show_announcement'
STATUS_ID = 'status_id'
SUITE_ID = 'suite_id'
SUITE_MODE = 'suite_mode'
START_ON = 'start_on'
TEMPLATE_ID = 'template_id'
TEST_ID = 'test_id'
TYPE_ID = 'type_id'
TITLE = 'title'
VERSION = 'version'

# Add/Delete/Update fields by object
CASE_ADD_FIELDS = (TITLE, TEMPLATE_ID, TYPE_ID, PRIORITY_ID, ESTIMATE,
                   MILESTONE_ID, REFS)
CASE_UPDATE_FIELDS = CASE_ADD_FIELDS
CONFIG_ADD_FIELDS = (NAME, )
CONFIG_UPDATE_FIELDS = CONFIG_ADD_FIELDS
CONFIG_GROUP_ADD_FIELDS = (NAME, )
CONFIG_GROUP_UPDATE_FIELDS = CONFIG_GROUP_ADD_FIELDS
MILESTONE_ADD_FIELDS = (NAME, DESCRIPTION, DUE_ON, PARENT_ID, START_ON)
# TODO: find out why `is_started` doesn't work, and add it back in
MILESTONE_UPDATE_FIELDS = MILESTONE_ADD_FIELDS + (IS_COMPLETED, IS_STARTED, START_ON)
PROJECT_ADD_FIELDS = (NAME, ANNOUNCEMENT, SHOW_ANNOUNCEMENT, SUITE_MODE)
PROJECT_UPDATE_FIELDS = PROJECT_ADD_FIELDS + (IS_COMPLETED, )
RESULT_ADD_FIELDS = (ASSIGNEDTO_ID, COMMENT, DEFECTS, ELAPSED, STATUS_ID, VERSION)
RUN_UPDATE_FIELDS = (NAME, DESCRIPTION, INCLUDE_ALL, CASE_IDS, MILESTONE_ID)
RUN_ADD_FIELDS = RUN_UPDATE_FIELDS + (SUITE_ID, ASSIGNEDTO_ID)
SECTION_ADD_FIELDS = (NAME, DESCRIPTION, PARENT_ID, SUITE_ID)
SECTION_UPDATE_FIELDS = (NAME, DESCRIPTION)
SUITE_ADD_FIELDS = (NAME, DESCRIPTION)
SUITE_UPDATE_FIELDS = SUITE_ADD_FIELDS



import traw


def main():
    client = traw.Client()  # Credentials loaded through ENV vars

    # Create a new project object
    new_project = client.project()
    new_project.name = "ACME Contract Testing"
    new_project.announcement = "For testing associated with the Acme Contract"
    new_project.show_announcement = True
    new_project.suite_mode = 1  # Single Suite mode

    # Add the project to testrail, replacing it with the Milestone returned
    # from the TestRail API.
    active_project = client.add(new_project)
    project_id = active_project.id

    # Verify some parameters
    assert active_project.name == "ACME Contract Testing"
    assert active_project.announcement == "For testing associated with the Acme Contract"
    assert active_project.is_completed is False

    # Lets say the project is now complete.
    # Update the name, announcement, and completion status
    active_project.name = "ACME Contract Testing - Complete"
    active_project.announcement = ("For testing associated with the Acme Contract,"
                                   "completed 2018/01/02")
    active_project.is_completed = True

    # POST the updated project to Testrail, replacing it with the Project returned
    # from the API
    closed_project = client.update(active_project)

    assert closed_project.name == "ACME Contract Testing - Complete"
    assert closed_project.announcement == ("For testing associated with the Acme Contract,"
                                           "completed 2018/01/02")
    assert closed_project.is_completed is True

    # Lets say its now a few years later, and we want to remove this old project
    # NOTE: your user must have admin privs to delete a project
    old_project = client.project(project_id)
    client.delete(old_project)


if __name__ == "__main__":
    main()



from copy import deepcopy

import pytest

from traw.models import Status

COLOR_BRIGHT = 16631751
COLOR_DARK = 14250867
COLOR_MEDIUM = 15829135
ID = 1234
LABEL = "Failed"
NAME = 'mock status name'


@pytest.fixture()
def status(client):
    content = {"color_bright": COLOR_BRIGHT,
               "color_dark": COLOR_DARK,
               "color_medium": COLOR_MEDIUM,
               "id": ID,
               "is_final": False,
               "is_system": True,
               "is_untested": False,
               "label": LABEL,
               "name": NAME}
    return Status(client, deepcopy(content))


def test_color_bright(status):
    assert status.color_bright == COLOR_BRIGHT


def test_color_dark(status):
    assert status.color_dark == COLOR_DARK


def test_color_medium(status):
    assert status.color_medium == COLOR_MEDIUM


def test_id(status):
    assert status.id == ID


def test_is_final(status):
    assert status.is_final is False


def test_is_system(status):
    assert status.is_system is True


def test_is_untested(status):
    assert status.is_untested is False


def test_label(status):
    assert status.label == LABEL


def test_name(status):
    assert status.name == NAME



from datetime import datetime as dt, timedelta

from .. import const
from .model_base import ModelBase
from .posters import Addable
from .status import Status
from .test import Test
from .user import User
from ..utils import duration_to_timedelta as dur_to_td


class Result(Addable, ModelBase):
    """ Object model for TestRail Results

    """
    _ADDABLE_FIELDS = const.RESULT_ADD_FIELDS

    @property
    def assigned_to(self):
        """ The assignee (user) of the test result """
        user_id = self._content.get('assignedto_id', None)
        return self.client.user(user_id) if user_id is not None else None

    @assigned_to.setter
    def assigned_to(self, val):
        if not isinstance(val, User):
            raise TypeError(const.SETTER_ERR.format(User, type(val)))
        self._content['assignedto_id'] = val.id

    @property
    def comment(self):
        """ The comment or error message of the test result """
        return self._content.get('comment', None)

    @comment.setter
    def comment(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['comment'] = val

    @property
    def created_by(self):
        """ The user who created the test result """
        user_id = self._content.get('created_by', None)
        return self.client.user(user_id) if user_id is not None else None

    @property
    def created_on(self):
        """ The date/time when the test result was created
            (as datetime.datetime object)
        """
        co = self._content.get('created_on', None)
        return dt.fromtimestamp(co) if co is not None else None

    @property
    def defects(self):
        """ Yields string defects that are linked to the test result """
        defs_raw = self._content.get('defects', None)
        defects = defs_raw.split(',') if defs_raw is not None else list()

        for defect in defects:
            yield defect

    @defects.setter
    def defects(self, val):
        """ Add defects associated with the test result

            Defects can be added as follows:
             - A defect string: ``result.defects = "DEFECT1"``
             - A comma separated string: ``result.defects = "DEFECT1,DEFECT2"
             - A list of defect strings: ["DEFECT1", "DEFECT2"]
        """
        if isinstance(val, str):
            defects_ = val
        elif isinstance(val, list) and all(map(lambda x: isinstance(x, str), val)):
            defects_ = ','.join(val)
        else:
            msg = ("Defects must be either a single string of one or more comma "
                   "separated defects ('DEFECT1' or 'DEFECT1,DEFECT2') or a list "
                   "of defect strings (['DEFECT1', 'DEFECT2']). Found {0}")
            raise TypeError(msg.format(val))

        self._content['defects'] = defects_

    @property
    def elapsed(self):
        """ The amount of time it took to execute the test e.g. "30s" or "1m 45s"
            (As a datetime.timedelta object)
        """
        elapsed = self._content.get('elapsed', None)
        return dur_to_td(elapsed) if elapsed is not None else None

    @elapsed.setter
    def elapsed(self, val):
        if not isinstance(val, timedelta):
            raise TypeError(const.SETTER_ERR.format(timedelta, type(val)))
        self._content['elapsed'] = val.seconds

    @property
    def status(self):
        """ The test status """
        status_id = self._content.get('status_id', None)
        return self.client.status(status_id) if status_id is not None else None

    @status.setter
    def status(self, val):
        if not isinstance(val, Status):
            raise TypeError(const.SETTER_ERR.format(Status, type(val)))
        self._content['status_id'] = val.id

    @property
    def test(self):
        """ The test this test result belongs to """
        test_id = self._content.get('test_id', None)
        return self.client.test(test_id) if test_id is not None else None

    @test.setter
    def test(self, val):
        if not isinstance(val, Test):
            raise TypeError(const.SETTER_ERR.format(Test, type(val)))
        self._content['test_id'] = val.id

    @property
    def version(self):
        """ The (build) version the test was executed against """
        return self._content.get('version', None)

    @version.setter
    def version(self, val):
        if not isinstance(val, str):
            raise TypeError(const.SETTER_ERR.format(str, type(val)))
        self._content['version'] = val



from .model_base import ModelBase
from ..utils import duration_to_timedelta as dur_to_td


class Test(ModelBase):
    """ Object model for TestRail Tests

    To get an existing test from the TestRail API:

    .. code-block:: python

        # Locate an existing test by test ID
        test = traw_client.test(1234)

    To get all tests for a run from the TestRail API:

    .. code-block:: python

        # Locate an existing run by run ID
        run = traw_client.run(4321)
        tests = traw_client.tests(run)

    """
    @property
    def assigned_to(self):
        """ The user the test is assigned to """
        user_id = self._content.get('assignedto_id', None)
        return self.client.user(user_id) if user_id is not None else None

    @property
    def case(self):
        """ The test case related to this test """
        return self.client.case(self._content['case_id'])

    @property
    def estimate(self):
        """ The estimated duration of the test, e.g. "30s" or "1m 45s" """
        est = self._content.get('estimate', None)
        return dur_to_td(est) if est is not None else None

    @property
    def estimate_forecast(self):
        """ The forecasted, estimated duration of the test, e.g. "30s" or "1m 45s" """
        est_forecast = self._content.get('estimate_forecast', None)
        return dur_to_td(est_forecast) if est_forecast is not None else None

    @property
    def milestone(self):  # TODO: find out why API returns only None
        """ The milestone that is linked to the test case """
        milestone_id = self._content.get('milestone_id', None)
        return self.client.milestone(milestone_id) if milestone_id is not None else None

    @property
    def priority(self):
        """ The priority that is linked to the test case """
        return self.client.priority(self._content['priority_id'])

    @property
    def refs(self):
        """ Yields string references/requirements that are linked to the test case """
        refs_raw = self._content.get('refs', None)
        refs = refs_raw.split(',') if refs_raw is not None else list()

        for ref in refs:
            yield ref

    @property
    def run(self):
        """ The test run that is linked to the test case """
        return self.client.run(self._content['run_id'])

    @property
    def status(self):
        """ The current test status of the test case """
        return self.client.status(self._content['status_id'])

    @property
    def title(self):
        """ The test title that is linked to the test case """
        return self._content['title']

    @property
    def type(self):
        """ The test case type that is linked to the test case """
        return self.client.case_type(self._content['type_id'])
